# -*- coding: utf-8 -*-

from datetime import datetime

from cf_units import Unit
import dask.array as da
import numpy as np

from .support import (
    mask_data_from_parameters,
    normalize_axis,
    IndexFunction,
    ThresholdMixin,
    ReducerMixin,
    RollingWindowMixin,
    DASK_OPERATORS,
)
import gordias.util.units


class CountLevelCrossings(IndexFunction):
    def __init__(self, threshold):
        super().__init__(units=Unit("1"))
        self.threshold = threshold
        self.extra_coords.append(threshold.copy())

    def prepare(self, input_cubes, parameters=None):
        props = {
            (cube.dtype, cube.units, cube.standard_name)
            for cube in input_cubes.values()
        }
        assert len(props) == 1
        dtype, units, standard_name = props.pop()
        threshold = self.threshold
        threshold.points = threshold.points.astype(dtype)
        if threshold.has_bounds():
            threshold.bounds = threshold.bounds.astype(dtype)
        gordias.util.units.change_units(threshold, units, standard_name)
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        cond = da.logical_and(
            data["low_data"] < self.threshold.points,
            self.threshold.points < data["high_data"],
        )
        res = np.count_nonzero(cond, axis=axis)
        return res.astype("float32")

    lazy_func = call_func


class CountOccurrences(ThresholdMixin, IndexFunction):
    def __init__(self, threshold, condition):
        super().__init__(threshold, condition, units=Unit("1"))

    def prepare(self, input_cubes, parameters=None):
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        cond = self.condition(data, self.threshold.points)
        res = np.count_nonzero(cond, axis=axis)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        cond = self.lazy_condition(data, self.threshold.points)
        res = da.count_nonzero(cond, axis=axis)
        return res.astype("float32")


class DiurnalTemperatureRange(ReducerMixin, IndexFunction):
    def __init__(self, statistic="mean"):
        super().__init__(statistic, units=Unit("degree_Celsius"))

    def prepare(self, input_cubes, parameters=None):
        props = {
            (cube.dtype, cube.units, cube.standard_name)
            for cube in input_cubes.values()
        }
        assert len(props) == 1
        dtype, units, standard_name = props.pop()
        assert units.is_convertible(Unit("degree_Celsius"))
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        res = self.reducer(data["high_data"] - data["low_data"], axis=axis)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        res = self.lazy_reducer(data["high_data"] - data["low_data"], axis=axis)
        return res.astype("float32")


class ExtremeTemperatureRange(IndexFunction):
    def __init__(self):
        super().__init__(units=Unit("degree_Celsius"))

    def prepare(self, input_cubes, parameters=None):
        props = {
            (cube.dtype, cube.units, cube.standard_name)
            for cube in input_cubes.values()
        }
        assert len(props) == 1
        dtype, units, standard_name = props.pop()
        assert units.is_convertible(Unit("degree_Celsius"))
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        res = data["high_data"].max(axis=axis) - data["low_data"].min(axis=axis)
        return res.astype("float32")

    lazy_func = call_func


class FirstOccurrence(ThresholdMixin, IndexFunction):
    def __init__(self, threshold, condition):
        super().__init__(threshold, condition, units=Unit("days"))

    def prepare(self, input_cubes, parameters=None):
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = np.ma.getmaskarray(data).any(axis=axis)
        cond = self.condition(data, self.threshold.points)
        res = np.where(cond.any(axis=axis), cond.argmax(axis=axis), np.nan)
        res = np.ma.masked_array(np.ma.getdata(res), mask)
        res = np.ma.masked_invalid(res)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        cond = self.lazy_condition(data, self.threshold.points)
        res = da.where(cond.any(axis=axis), cond.argmax(axis=axis), np.nan)
        res = da.ma.masked_array(da.ma.getdata(res), mask)
        res = da.ma.masked_invalid(res)
        return res.astype("float32")

    def post_process(self, cube, data, coords, period, **kwargs):
        time = cube.coord("time")
        calendar = time.units.calendar
        offsets = np.empty_like(time.points, dtype=data.dtype)
        for i, representative_date in enumerate(time.cells()):
            year = representative_date.point.year
            start_date = datetime(year, period.first_month_number, 1)
            units = Unit(f"days since {year}-01-01", calendar=calendar)
            offsets[i] = units.date2num(start_date)
        result_data = data + offsets[:, None, None]
        return cube, result_data


class InterdayDiurnalTemperatureRange(IndexFunction):
    def __init__(self):
        super().__init__(units=Unit("degree_Celsius"))

    def prepare(self, input_cubes, parameters=None):
        props = {
            (cube.dtype, cube.units, cube.standard_name)
            for cube in input_cubes.values()
        }
        assert len(props) == 1
        dtype, units, standard_name = props.pop()
        assert units.is_convertible(Unit("degree_Celsius"))
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        res = np.absolute(
            np.diff(data["high_data"] - data["low_data"], axis=axis)
        ).mean(axis=axis)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        res = da.absolute(
            da.diff(data["high_data"] - data["low_data"], axis=axis)
        ).mean(axis=axis)
        return res.astype("float32")


class LastOccurrence(ThresholdMixin, IndexFunction):
    def __init__(self, threshold, condition):
        super().__init__(threshold, condition, units=Unit("days"))

    def prepare(self, input_cubes, parameters=None):
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        super().prepare(input_cubes)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = np.ma.getmaskarray(data).any(axis=axis)
        cond = self.condition(np.flip(data, axis=axis), self.threshold.points)
        ndays = data.shape[axis]
        res = np.where(cond.any(axis=axis), ndays - cond.argmax(axis=axis), np.nan)
        res = np.ma.masked_array(np.ma.getdata(res), mask)
        res = np.ma.masked_invalid(res)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        cond = self.lazy_condition(da.flip(data, axis), self.threshold.points)
        ndays = data.shape[axis]
        res = da.where(cond.any(axis=axis), ndays - cond.argmax(axis=axis), np.nan)
        res = da.ma.masked_array(da.ma.getdata(res), mask)
        res = da.ma.masked_invalid(res)
        return res.astype("float32")

    def post_process(self, cube, data, coords, period, **kwargs):
        time = cube.coord("time")
        calendar = time.units.calendar
        offsets = np.empty_like(time.points, dtype=data.dtype)
        for i, representative_date in enumerate(time.cells()):
            year = representative_date.point.year
            start_date = datetime(year, period.first_month_number, 1)
            units = Unit(f"days since {year}-01-01", calendar=calendar)
            offsets[i] = units.date2num(start_date)
        result_data = data + offsets[:, None, None]
        return cube, result_data


class Percentile(IndexFunction):
    def __init__(self, percentiles, method="linear"):
        super().__init__()
        self.percentiles = percentiles
        assert np.all(self.percentiles.points > 0)
        assert np.all(self.percentiles.points < 100)
        self.method = method

    def prepare(self, input_cubes, parameters=None):
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name
        self.units = ref_cube.units
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = np.ma.getmaskarray(data).any(axis=axis)
        filled = np.ma.filled(data, fill_value=np.nan)
        res = np.nanpercentile(
            filled, q=self.percentiles.points, axis=axis, method=self.method
        ).squeeze()
        res = np.ma.masked_array(np.ma.getdata(res), mask)
        res = np.ma.masked_where(np.isnan(res), res)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        filled = da.ma.filled(data, fill_value=np.nan)

        def percentile(arr):
            return np.nanpercentile(arr, q=self.percentiles.points, method=self.method)

        res = da.apply_along_axis(percentile, axis=axis, arr=filled).squeeze()
        res = da.ma.masked_array(da.ma.getdata(res), mask)
        res = da.ma.masked_where(da.isnan(res), res)
        return res.astype("float32")


class Statistics(ReducerMixin, IndexFunction):
    def __init__(self, statistic):
        super().__init__(statistic)

    def prepare(self, input_cubes, parameters=None):
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name
        self.units = ref_cube.units
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        res = self.reducer(data, axis=axis)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        res = self.lazy_reducer(data, axis=axis)
        return res.astype("float32")


class ThresholdedPercentile(ThresholdMixin, IndexFunction):
    def __init__(self, threshold, condition, percentiles, method="linear"):
        super().__init__(threshold, condition)
        self.percentiles = percentiles
        assert np.all(self.percentiles.points > 0)
        assert np.all(self.percentiles.points < 100)
        self.method = method

    def prepare(self, input_cubes, parameters=None):
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name
        self.units = ref_cube.units
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = np.ma.getmaskarray(data).any(axis=axis)
        cond = self.condition(data, self.threshold.points)
        masked = np.ma.masked_where(~cond, data)
        filled = np.ma.filled(masked, fill_value=np.nan)
        res = np.nanpercentile(
            filled,
            q=self.percentiles.points,
            axis=axis,
            method=self.method,
        ).squeeze()
        res = np.ma.masked_array(np.ma.getdata(res), mask)
        res = np.ma.masked_where(np.isnan(res), res)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        cond = self.condition(data, self.threshold.points)
        masked = da.ma.masked_where(~cond, data)
        filled = da.ma.filled(masked, fill_value=np.nan)

        def percentile(arr):
            return np.nanpercentile(arr, q=self.percentiles.points, method=self.method)

        res = da.apply_along_axis(percentile, axis=axis, arr=filled).squeeze()
        res = da.ma.masked_array(da.ma.getdata(res), mask)
        res = da.ma.masked_where(da.isnan(res), res)
        return res.astype("float32")


class ThresholdedStatistics(ThresholdMixin, ReducerMixin, IndexFunction):
    def __init__(self, threshold, condition, statistic):
        super().__init__(threshold, condition, statistic, units=Unit("days"))

    def prepare(self, input_cubes, parameters=None):
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name
        self.units = ref_cube.units
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        comb = self.condition(data, self.threshold.points)
        res = self.reducer(np.ma.masked_where(~comb, data), axis=axis)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        comb = self.lazy_condition(data, self.threshold.points)
        res = self.lazy_reducer(da.ma.masked_where(~comb, data), axis=axis)
        return res.astype("float32")


class RunningStatistics(RollingWindowMixin, IndexFunction):
    def __init__(self, rolling_aggregator, window_size, overall_statistic):
        super().__init__(rolling_aggregator, window_size, overall_statistic)
        self.fuse_periods = True
        self.bandwidth = self.window_size.points[0] // 2
        self.tail_overlap = self.window_size.points[0] - 1
        self.head_overlap = self.tail_overlap + self.window_size.points[0] % 2

    def prepare(self, input_cubes, parameters=None):
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name
        self.units = ref_cube.units
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )

    def pre_aggregate_shape(self, *args, **kwargs):
        return (self.head_overlap + self.tail_overlap + 1,)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = np.ma.getmaskarray(data).any(axis=axis)
        rolling_view = np.lib.stride_tricks.sliding_window_view(
            data, self.window_size.points, axis
        )
        aggregated = self.rolling_aggregator(rolling_view, -1)
        reduced = self.reducer(aggregated, axis=axis)
        masked = np.ma.masked_array(np.ma.getdata(reduced), mask)
        head_slices = (slice(None, None),) * axis + (slice(None, self.head_overlap),)
        head = np.moveaxis(data[head_slices], axis, -1)
        tail_slices = (slice(None, None),) * axis + (slice(-self.tail_overlap, None),)
        tail = np.moveaxis(data[tail_slices], axis, -1)
        res = np.ma.concatenate([head, masked[..., np.newaxis], tail], axis=-1)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        rolling_view = da.overlap.sliding_window_view(
            data, self.window_size.points, axis
        )
        aggregated = self.lazy_rolling_aggregator(rolling_view, -1)
        reduced = self.lazy_reducer(aggregated, axis=axis)
        masked = da.ma.masked_array(da.ma.getdata(reduced), mask)
        head_slices = (slice(None, None),) * axis + (slice(None, self.head_overlap),)
        head = da.moveaxis(data[head_slices], axis, -1)
        tail_slices = (slice(None, None),) * axis + (slice(-self.tail_overlap, None),)
        tail = da.moveaxis(data[tail_slices], axis, -1)
        res = da.concatenate([head, masked[..., np.newaxis], tail], axis=-1)
        return res.astype("float32")

    def post_process(self, cube, data, coords, period, **kwargs):
        def _aggregate(tail, head):
            window_size = self.window_size.points
            overlap = da.concatenate([tail, head], axis=-1)
            rolling_view = da.overlap.sliding_window_view(overlap, window_size, -1)
            aggregated = self.lazy_rolling_aggregator(rolling_view, axis=-1)
            isnan = da.isnan(overlap)
            if isnan.any():
                rolling_isnan = da.overlap.sliding_window_view(isnan, window_size, -1)
                mask = da.any(rolling_isnan, axis=-1)
                aggregated = da.ma.masked_where(mask, aggregated)
            return aggregated

        def _get_strict_mask(head, pre_stat, tail):
            head_mask = da.ma.getmaskarray(head).any(axis=-1)
            pre_stat_mask = da.ma.getmaskarray(pre_stat)
            tail_mask = da.ma.getmaskarray(tail).any(axis=-1)
            strict_mask = head_mask | pre_stat_mask | tail_mask
            return strict_mask

        def fuse(this, previous_tail, next_head):
            head = this[..., : self.head_overlap]
            pre_statistic = this[..., self.head_overlap]
            tail = this[..., -self.tail_overlap :].copy()
            strict_mask = _get_strict_mask(next_head, pre_statistic, previous_tail)
            head_aggregated = _aggregate(previous_tail[..., -self.bandwidth :], head)
            tail_aggregated = _aggregate(tail, next_head[..., : self.bandwidth])
            concatenated = da.concatenate(
                [head_aggregated, pre_statistic[..., np.newaxis], tail_aggregated],
                axis=-1,
            )
            running_statistic = self.lazy_reducer(concatenated, axis=-1)
            masked = da.ma.masked_where(strict_mask, running_statistic)
            return masked, tail

        if self.fuse_periods and len(data) > 1:
            stack = []
            this = data[0]
            tail_shape = this.shape[:-1] + (self.tail_overlap,)
            previous_tail = da.full(tail_shape, fill_value=np.nan, dtype=np.float32)

            for next_chunk in data[1:]:
                next_head = next_chunk[..., : self.head_overlap].copy()
                running_statistic, previous_tail = fuse(this, previous_tail, next_head)
                stack.append(running_statistic)
                this = next_chunk

            head_shape = this.shape[:-1] + (self.head_overlap,)
            next_head = da.full(head_shape, fill_value=np.nan, dtype=np.float32)
            stack.append(fuse(next_chunk, previous_tail, next_head)[0])
            res_data = da.stack(stack, axis=0)
        else:
            stack = []
            for this in data:
                stack.append(this[..., self.head_overlap])
            res_data = da.stack(stack, axis=0)
        return cube, res_data


class ThresholdedRunningStatistics(ThresholdMixin, RunningStatistics):
    def __init__(
        self, threshold, condition, rolling_aggregator, window_size, overall_statistic
    ):
        super().__init__(
            threshold, condition, rolling_aggregator, window_size, overall_statistic
        )

    def call_func(self, data, axis, **kwargs):
        comb = self.condition(data, self.threshold.points)
        thresholded_data = np.ma.where(comb, data, 0.0)
        return super().call_func(thresholded_data, axis, **kwargs)

    def lazy_func(self, data, axis, **kwargs):
        comb = self.condition(data, self.threshold.points)
        thresholded_data = da.ma.where(comb, data, 0.0)
        return super().lazy_func(thresholded_data, axis, **kwargs)


class TemperatureSum(ThresholdMixin, IndexFunction):
    def __init__(self, threshold, condition):
        super().__init__(threshold, condition, units=Unit("days"))
        if condition in [">", ">="]:
            self.fun = lambda d, t: np.maximum(d - t, 0)
            self.lazy_fun = lambda d, t: da.maximum(d - t, 0)
        else:
            self.fun = lambda d, t: np.maximum(t - d, 0)
            self.lazy_fun = lambda d, t: da.maximum(t - d, 0)

    def prepare(self, input_cubes, parameters=None):
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name
        if ref_cube.units.is_convertible("degC"):
            self.units = "degC days"
        else:
            raise RuntimeError("Invalid input units")
        if parameters is not None:
            aux_coords = mask_data_from_parameters(input_cubes, parameters)
            self.extra_coords.extend(aux_coords)

    def call_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        threshold = self.threshold.points[0]
        res = np.sum(self.fun(data, threshold), axis=axis)
        return res.astype("float32")

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        threshold = self.threshold.points[0]
        res = da.sum(self.lazy_fun(data, threshold), axis=axis)
        return res.astype("float32")


class CountJointOccurrences(IndexFunction):
    def __init__(self, mapping):
        super().__init__(units=Unit("1"))
        self.mapping = mapping

    def add_extra_coords(self, input_cubes):
        thresholds = []
        for name, _ in input_cubes.items():
            for threshold, _ in self.mapping[name]:
                thresholds.append(threshold)
        if (
            thresholds[0].metadata.equal(thresholds[1].metadata, lenient=True)
            and thresholds[0].metadata.long_name == thresholds[1].metadata.long_name
        ):
            raise ValueError(
                "Auxiliary coordinates with the same metadata must have different "
                "<long_name>."
            )
        self.extra_coords.append(thresholds[0].copy())
        self.extra_coords.append(thresholds[1].copy())

    def prepare(self, input_cubes, parameters=None):
        self.add_extra_coords(input_cubes)
        props = {
            (name, cube.dtype, cube.units, cube.standard_name)
            for name, cube in input_cubes.items()
        }
        for data_name, dtype, units, standard_name in props:
            for threshold, _ in self.mapping[data_name]:
                threshold.points = threshold.points.astype(dtype)
                if threshold.has_bounds():
                    threshold.bounds = threshold.bounds.astype(dtype)
                gordias.util.units.change_units(threshold, units, standard_name)
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        super().prepare(input_cubes)

    def lazy_func(self, data, axis, **kwargs):
        if not isinstance(data, dict):
            data = {"data": data}
        conditions = []
        mask = []
        for name, arr in data.items():
            mask.append(da.ma.getmaskarray(arr).any(axis=axis))
            for threshold, operator in self.mapping[name]:
                conditions.append(operator(arr, threshold.points))
        filled = da.ma.filled(da.stack(conditions, axis=0), fill_value=False)
        cond = da.all(filled, axis=0)
        res = da.count_nonzero(cond, axis=axis)
        masked = da.ma.masked_array(
            da.ma.getdata(res), da.stack(mask, axis=0).any(axis=0)
        )
        return masked.astype("float32")

    call_func = lazy_func


class CountJointOccurrencesPrecipitationTemperature(CountJointOccurrences):
    def __init__(
        self,
        threshold_precip_data,
        threshold_temp_data,
        condition_precip_data,
        condition_temp_data,
    ):
        super().__init__(
            mapping={
                "precip_data": [
                    (threshold_precip_data, DASK_OPERATORS[condition_precip_data])
                ],
                "temp_data": [
                    (threshold_temp_data, DASK_OPERATORS[condition_temp_data])
                ],
            }
        )


class CountJointOccurrencesTemperature(CountJointOccurrences):
    def __init__(
        self,
        threshold_low_data,
        threshold_high_data,
        condition_low_data,
        condition_high_data,
    ):
        super().__init__(
            mapping={
                "data": [
                    (threshold_low_data, DASK_OPERATORS[condition_low_data]),
                    (threshold_high_data, DASK_OPERATORS[condition_high_data]),
                ]
            }
        )


class CountJointOccurrencesPrecipitationDoubleTemperature(CountJointOccurrences):
    def __init__(
        self,
        threshold_precip_data,
        threshold_temp_low_data,
        threshold_temp_high_data,
        condition_precip_data,
        condition_temp_low_data,
        condition_temp_high_data,
    ):
        super().__init__(
            mapping={
                "precip_data": [
                    (threshold_precip_data, DASK_OPERATORS[condition_precip_data])
                ],
                "temp_data": [
                    (threshold_temp_low_data, DASK_OPERATORS[condition_temp_low_data]),
                    (
                        threshold_temp_high_data,
                        DASK_OPERATORS[condition_temp_high_data],
                    ),
                ],
            }
        )



from contextlib import nullcontext as does_not_raise
import climix.index_functions.support
import numpy as np
import pytest
import iris.cube
import iris.coords
import iris.coord_categorisation
from typing import Any
import dask.array as da
from cf_units import Unit

TEST_ADD_MASK: dict[str, Any] = {
    "mask-start": (
        np.array(np.arange(8)).reshape(8, 1, 1),
        np.array([3]).reshape(1, 1),
        8,
        "start",
        does_not_raise(
            np.ma.masked_array(
                data=np.array(np.arange(8)).reshape(8, 1, 1),
                mask=np.array([[1, 1, 0, 0, 0, 0, 0, 0]]),
            )
        ),
    ),
    "mask-end": (
        np.array(np.arange(8)).reshape(8, 1, 1),
        np.array([3]).reshape(1, 1),
        8,
        "end",
        does_not_raise(
            np.ma.masked_array(
                data=np.array(np.arange(8)).reshape(8, 1, 1),
                mask=np.array([[0, 0, 1, 1, 1, 1, 1, 1]]),
            )
        ),
    ),
    "mask-end-threshold-1": (
        np.array(np.arange(8)).reshape(8, 1, 1),
        np.array([1]).reshape(1, 1),
        8,
        "end",
        does_not_raise(
            np.ma.masked_array(
                data=np.array(np.arange(8)).reshape(8, 1, 1),
                mask=np.array([[1, 1, 1, 1, 1, 1, 1, 1]]),
            )
        ),
    ),
    "mask-start-threshold-1": (
        np.array(np.arange(8)).reshape(8, 1, 1),
        np.array([1]).reshape(1, 1),
        8,
        "start",
        does_not_raise(
            np.ma.masked_array(
                data=np.array(np.arange(8)).reshape(8, 1, 1),
                mask=np.array([[0, 0, 0, 0, 0, 0, 0, 0]]),
            )
        ),
    ),
    "nan-values": (
        np.array(np.arange(8)).reshape(8, 1, 1),
        np.array([np.nan]).reshape(1, 1),
        8,
        "start",
        does_not_raise(
            np.ma.masked_array(
                data=np.array(np.arange(8)).reshape(8, 1, 1),
                mask=np.array([[1, 1, 1, 1, 1, 1, 1, 1]]),
            )
        ),
    ),
    "2x2": (
        np.array(np.arange(40)).reshape(10, 2, 2),
        np.array([[1, 3], [4, 10]]),
        10,
        "start",
        does_not_raise(
            np.ma.masked_array(
                data=np.array(np.arange(40)).reshape(10, 2, 2),
                mask=np.array(
                    [
                        [[0, 1], [1, 1]],
                        [[0, 1], [1, 1]],
                        [[0, 0], [1, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 0]],
                    ]
                ),
            )
        ),
    ),
}


parameter_names: tuple[str, ...] = (
    "data",
    "parameter_data",
    "length",
    "time_constraint",
    "expected_output",
)


@pytest.mark.parametrize(
    parameter_names,
    TEST_ADD_MASK.values(),
    ids=TEST_ADD_MASK.keys(),
)
def test_add_mask(data, parameter_data, length, time_constraint, expected_output):
    """Test funtion _add_mask()."""
    with expected_output:
        res = climix.index_functions.support._add_mask(
            data, parameter_data, length, time_constraint
        )
        assert np.array_equal(res, expected_output.enter_result, equal_nan=True)
        expected_mask = da.ma.getmaskarray(expected_output.enter_result)
        res_mask = da.ma.getmaskarray(res)
        assert (res_mask == expected_mask).all()
        assert res.dtype == data.dtype


TEST_MASK_DATA: dict[str, Any] = {
    "mask-start": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.arange(40).reshape(10, 2, 2),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 10, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=1, stop=9).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "start",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.arange(40).reshape(10, 2, 2),
                    mask=[
                        [[0, 1], [1, 1]],
                        [[0, 0], [1, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                    ],
                ),
                "parameters": np.arange(start=1, stop=5).reshape(1, 2, 2),
            }
        ),
    ),
    "mask-end": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(np.arange(40).reshape(10, 2, 2)).rechunk(
                            1, 1, 1
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 10, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=1, stop=9).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "end",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.arange(40).reshape(10, 2, 2),
                    mask=[
                        [[1, 0], [0, 0]],
                        [[1, 1], [0, 0]],
                        [[1, 1], [1, 0]],
                        [[1, 1], [1, 1]],
                        [[1, 1], [1, 1]],
                        [[1, 1], [1, 1]],
                        [[1, 1], [1, 1]],
                        [[1, 1], [1, 1]],
                        [[1, 1], [1, 1]],
                        [[1, 1], [1, 1]],
                    ],
                ),
                "parameters": np.arange(start=1, stop=5).reshape(1, 2, 2),
            }
        ),
    ),
    "nan-values": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.array(
                            [
                                [[0, 1], [np.nan, 2]],
                                [[0, 1], [3, 2]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, 2]],
                            ],
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=2, stop=10).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "end",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=[
                        [[0, 1], [np.nan, 2]],
                        [[0, 1], [3, 2]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, 2]],
                    ],
                    mask=[
                        [[0, 0], [0, 0]],
                        [[1, 0], [0, 0]],
                        [[1, 1], [0, 0]],
                        [[1, 1], [1, 0]],
                        [[1, 1], [1, 1]],
                    ],
                ),
                "parameters": np.arange(start=2, stop=6).reshape(1, 2, 2),
            }
        ),
    ),
    "masked-input": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.ma.masked_array(
                            data=[
                                [[0, 1], [np.nan, 2]],
                                [[0, 1], [3, 2]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, 2]],
                            ],
                            mask=[
                                [[0, 1], [0, 1]],
                                [[0, 0], [0, 0]],
                                [[0, 0], [0, 0]],
                                [[0, 0], [0, 0]],
                                [[0, 0], [0, 1]],
                            ],
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=2, stop=10).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "end",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=[
                        [[0, 1], [np.nan, 2]],
                        [[0, 1], [3, 2]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, 2]],
                    ],
                    mask=[
                        [[0, 1], [0, 1]],
                        [[1, 0], [0, 0]],
                        [[1, 1], [0, 0]],
                        [[1, 1], [1, 0]],
                        [[1, 1], [1, 1]],
                    ],
                ),
                "parameters": np.arange(start=2, stop=6).reshape(1, 2, 2),
            }
        ),
    ),
    "nan-values-in-parameters": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.ma.masked_array(
                            data=[
                                [[0, 1], [np.nan, 2]],
                                [[0, 1], [3, 2]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, 2]],
                            ],
                            mask=[
                                [[0, 1], [0, 1]],
                                [[0, 0], [0, 0]],
                                [[0, 0], [0, 0]],
                                [[0, 0], [0, 0]],
                                [[0, 0], [0, 1]],
                            ],
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        np.array(
                            [
                                [[np.nan, 1], [3, 2]],
                                [[0, np.nan], [3, 2]],
                            ]
                        ),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "start",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=[
                        [[0, 1], [np.nan, 2]],
                        [[0, 1], [3, 2]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, 2]],
                    ],
                    mask=[
                        [[1, 1], [1, 1]],
                        [[1, 0], [1, 0]],
                        [[1, 0], [0, 0]],
                        [[1, 0], [0, 0]],
                        [[1, 0], [0, 1]],
                    ],
                ),
                "parameters": np.array(
                    [
                        [[np.nan, 1], [3, 2]],
                    ]
                ),
            }
        ),
    ),
    "high-threshold-values-start": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.array(
                            [
                                [[0, 1], [np.nan, 2]],
                                [[0, 1], [3, 2]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, 2]],
                            ],
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        np.array(
                            [
                                [[365, 1], [182, 3]],
                                [[365, 1], [182, 3]],
                            ]
                        ),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "start",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=[
                        [[0, 1], [np.nan, 2]],
                        [[0, 1], [3, 2]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, 2]],
                    ],
                    mask=[
                        [[1, 0], [1, 1]],
                        [[1, 0], [1, 1]],
                        [[1, 0], [1, 0]],
                        [[1, 0], [1, 0]],
                        [[1, 0], [1, 0]],
                    ],
                ),
                "parameters": np.array(
                    [
                        [[365, 1], [182, 3]],
                    ]
                ),
            }
        ),
    ),
    "high-threshold-values-end": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.array(
                            [
                                [[0, 1], [np.nan, 2]],
                                [[0, 1], [3, 2]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, np.nan]],
                                [[0, 1], [3, 2]],
                            ],
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        np.array(
                            [
                                [[365, 1], [182, 3]],
                                [[365, 1], [182, 3]],
                            ]
                        ),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "end",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=[
                        [[0, 1], [np.nan, 2]],
                        [[0, 1], [3, 2]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, np.nan]],
                        [[0, 1], [3, 2]],
                    ],
                    mask=[
                        [[0, 1], [0, 0]],
                        [[0, 1], [0, 0]],
                        [[0, 1], [0, 1]],
                        [[0, 1], [0, 1]],
                        [[0, 1], [0, 1]],
                    ],
                ),
                "parameters": np.array(
                    [
                        [[365, 1], [182, 3]],
                    ]
                ),
            }
        ),
    ),
    "mask_start-two-years": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.arange(730).reshape(730, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 730, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=183, stop=185).reshape(2, 1, 1),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "start",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.arange(730).reshape(730, 1, 1),
                    mask=np.concatenate(
                        (np.ones(182), np.zeros(183), np.ones(183), np.zeros(182)),
                        axis=0,
                    ),
                ),
                "parameters": np.arange(start=183, stop=185).reshape(2, 1, 1),
            }
        ),
    ),
    "mask-start-offset": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.ones(362).reshape(362, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.concatenate(
                            (np.arange(31, 212, 1), np.arange(396, 577, 1)),
                            axis=0,
                        ),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.array([32, 182]).reshape(2, 1, 1),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "start",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.ones(362).reshape(362, 1, 1),
                    mask=np.concatenate(
                        (np.zeros(181), np.ones(150), np.zeros(31)),
                        axis=0,
                    ).reshape(362, 1, 1),
                ),
                "parameters": np.array([1, 151]).reshape(2, 1, 1),
            }
        ),
    ),
    "mask-end-offset": (
        {
            "cubes": {
                "cube_1": {
                    "cube": iris.cube.Cube(
                        data=np.ones(362).reshape(362, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.concatenate(
                            (np.arange(31, 212, 1), np.arange(396, 577, 1)),
                            axis=0,
                        ),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "cube_2": {
                    "cube": iris.cube.Cube(
                        data=np.array([31, 182]).reshape(2, 1, 1),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        "end",
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.ones(362).reshape(362, 1, 1),
                    mask=np.concatenate(
                        (np.ones(181), np.zeros(150), np.ones(31)),
                        axis=0,
                    ).reshape(362, 1, 1),
                ),
                "parameters": np.array([0, 151]).reshape(2, 1, 1),
            }
        ),
    ),
}


parameter_names = (
    "f_cubes",
    "time_constraint",
    "expected_output",
)
fixtures = ["f_cubes"]


@pytest.mark.parametrize(
    parameter_names,
    TEST_MASK_DATA.values(),
    ids=TEST_MASK_DATA.keys(),
    indirect=fixtures,
)
def test_mask_data(f_cubes, time_constraint, expected_output):
    """Test funtion _mask_data()."""
    with expected_output:
        cube_1 = f_cubes["cube_1"]
        cube_2 = f_cubes["cube_2"]
        iris.coord_categorisation.add_year(cube_1, "time")
        iris.coord_categorisation.add_year(cube_2, "time")
        res, parameters = climix.index_functions.support._mask_data(
            cube_1, cube_2, time_constraint
        )
        expected_result = expected_output.enter_result["data"]
        expected_parameters = expected_output.enter_result["parameters"]
        assert np.array_equal(res, expected_result, equal_nan=True)
        expected_mask = da.ma.getmaskarray(expected_result)
        res_mask = da.ma.getmaskarray(res)
        assert (res_mask == expected_mask).all()
        assert res.dtype == np.float32
        assert np.array_equal(parameters, expected_parameters, equal_nan=True)


TEST_MASK_DATA_FROM_PARAMETERS: dict[str, Any] = {
    "mask-start-and-end": (
        {
            "cubes": {
                "data": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(np.arange(40).reshape(10, 2, 2)).rechunk(
                            1, 1, 1
                        ),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 10, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "start": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=1, stop=9).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "end": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=8, stop=16).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.arange(40).reshape(10, 2, 2),
                    mask=[
                        [[0, 1], [1, 1]],
                        [[0, 0], [1, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[1, 0], [0, 0]],
                        [[1, 1], [0, 0]],
                        [[1, 1], [1, 0]],
                    ],
                    dtype=np.int64,
                ),
                "aux_coord": [
                    iris.coords.AuxCoord(
                        np.arange(start=1, stop=5).reshape(1, 2, 2),
                        var_name="mask-start",
                        long_name="Day-of-year for which data have been masked.",
                        units=Unit("day"),
                    ),
                    iris.coords.AuxCoord(
                        np.arange(start=8, stop=12).reshape(1, 2, 2),
                        var_name="mask-end",
                        long_name="Day-of-year for which data have been masked.",
                        units=Unit("day"),
                    ),
                ],
            },
        ),
    ),
    "mask-start": (
        {
            "cubes": {
                "data": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(
                            np.arange(40, dtype=np.int32).reshape(10, 2, 2)
                        ).rechunk(1, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 10, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "start": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=1, stop=9).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        does_not_raise(
            {
                "data": np.ma.masked_array(
                    data=np.arange(40).reshape(10, 2, 2),
                    mask=[
                        [[0, 1], [1, 1]],
                        [[0, 0], [1, 1]],
                        [[0, 0], [0, 1]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                        [[0, 0], [0, 0]],
                    ],
                    dtype=np.int32,
                ),
                "aux_coord": [
                    iris.coords.AuxCoord(
                        np.arange(start=1, stop=5).reshape(1, 2, 2),
                        var_name="mask-start",
                        long_name="Day-of-year for which data have been masked.",
                        units=Unit("day"),
                    )
                ],
            },
        ),
    ),
    "mask-end": (
        {
            "cubes": {
                "data": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(
                            np.arange(40, dtype=np.float32).reshape(10, 2, 2)
                        ).rechunk(1, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0, 10, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
                "end": {
                    "cube": iris.cube.Cube(
                        data=np.arange(start=8, stop=16).reshape(2, 2, 2),
                        standard_name=None,
                        var_name=None,
                        units="day",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.array([183, 365.5]),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            }
        },
        does_not_raise(
            {
                "data": (
                    np.ma.masked_array(
                        data=np.arange(40).reshape(10, 2, 2),
                        mask=[
                            [[0, 0], [0, 0]],
                            [[0, 0], [0, 0]],
                            [[0, 0], [0, 0]],
                            [[0, 0], [0, 0]],
                            [[0, 0], [0, 0]],
                            [[0, 0], [0, 0]],
                            [[0, 0], [0, 0]],
                            [[1, 0], [0, 0]],
                            [[1, 1], [0, 0]],
                            [[1, 1], [1, 0]],
                        ],
                        dtype=np.float32,
                    )
                ),
                "aux_coord": [
                    iris.coords.AuxCoord(
                        np.arange(start=8, stop=12).reshape(1, 2, 2),
                        var_name="mask-end",
                        long_name="Day-of-year for which data have been masked.",
                        units=Unit("day"),
                    )
                ],
            }
        ),
    ),
}


parameter_names = ("f_cubes", "expected_output")
fixtures = ["f_cubes"]


@pytest.mark.parametrize(
    parameter_names,
    TEST_MASK_DATA_FROM_PARAMETERS.values(),
    ids=TEST_MASK_DATA_FROM_PARAMETERS.keys(),
    indirect=fixtures,
)
def test_mask_data_from_parameters(f_cubes, expected_output):
    """Test funtion mask_data_from_parameters()."""
    with expected_output:
        params = {}
        for key, value in f_cubes.items():
            iris.coord_categorisation.add_year(value, "time")
            params.update({key: value})
        input_cube = params.pop("data")
        dtype = input_cube.data.dtype
        input_cubes = {"data": input_cube}
        aux_coord = climix.index_functions.support.mask_data_from_parameters(
            input_cubes, params
        )
        res = input_cubes["data"].data
        expected_res = expected_output.enter_result["data"]
        expected_aux_coord = expected_output.enter_result["aux_coord"]
        assert np.array_equal(res, expected_res, equal_nan=True)
        expected_mask = da.ma.getmaskarray(expected_res)
        res_mask = da.ma.getmaskarray(res)
        assert (res_mask == expected_mask).all()
        assert res.dtype == dtype == expected_res.dtype
        for coord, exp in zip(aux_coord, expected_aux_coord):
            assert coord.var_name == exp.var_name
            assert coord.long_name == exp.long_name
            assert np.array_equal(coord.points, exp.points, equal_nan=True)
            assert coord.units == exp.units


TEST_COMPUTE_OFFSET: dict[str, Any] = {
    "offset-0": (
        {
            "cubes": {
                "cube": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(
                            np.arange(40, dtype=np.float32).reshape(10, 2, 2)
                        ).rechunk(1, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(0.5, 10.5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            },
        },
        2022,
        does_not_raise(0),
    ),
    "offset-31": (
        {
            "cubes": {
                "cube": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(
                            np.arange(40, dtype=np.float32).reshape(10, 2, 2)
                        ).rechunk(1, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(31.5, 41.5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            },
        },
        2022,
        does_not_raise(31),
    ),
    "offset-100": (
        {
            "cubes": {
                "cube": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(
                            np.arange(40, dtype=np.float32).reshape(10, 2, 2)
                        ).rechunk(1, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(100.5, 110.5, 1),
                        var_name="time",
                        standard_name="time",
                        units="days since 2022-01-01 00:00:00",
                    ),
                },
            },
        },
        2022,
        does_not_raise(100),
    ),
    "offset-seconds-31-days": (
        {
            "cubes": {
                "cube": {
                    "cube": iris.cube.Cube(
                        data=da.asarray(
                            np.arange(40, dtype=np.float32).reshape(10, 2, 2)
                        ).rechunk(1, 1, 1),
                        standard_name="air_temperature",
                        var_name="tas",
                        units="degree_Celsius",
                    ),
                    "dim_coord_time": iris.coords.DimCoord(
                        points=np.arange(2721600, 3585600, 86400),
                        var_name="time",
                        standard_name="time",
                        units="seconds since 2022-01-01 00:00:00",
                    ),
                },
            },
        },
        2022,
        does_not_raise(31),
    ),
}


parameter_names = ("f_cubes", "year", "expected_output")
fixtures = ["f_cubes"]


@pytest.mark.parametrize(
    parameter_names,
    TEST_COMPUTE_OFFSET.values(),
    ids=TEST_COMPUTE_OFFSET.keys(),
    indirect=fixtures,
)
def test_compute_offset(f_cubes, year, expected_output):
    """Test funtion _compute_offset()."""
    with expected_output:
        cube = f_cubes["cube"]
        offset = climix.index_functions.support._compute_offset(cube, year)
        assert offset == expected_output.enter_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import glob
import logging
import os
import textwrap
import time

import gordias.dask_setup
import gordias.datahandling
import gordias.util.cmip_path


from climix import __version__
from climix.metadata import load_metadata
from climix.period import Annual

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def parse_args():
    parser = argparse.ArgumentParser(
        description=(f"A climate index package, version {__version__}."),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-l",
        "--log-level",
        choices=LOG_LEVELS.keys(),
        default="info",
        help="the lowest priority level of log messages to display",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="write more detailed log messages"
    )
    parser.add_argument("-d", "--dask-scheduler", default="distributed-local-cluster")
    parser.add_argument(
        "-k",
        "--keep-open",
        action="store_true",
        help="keep climix running until key press (useful for debugging)",
    )
    parser.add_argument(
        "-p",
        "--period",
        help=textwrap.dedent(
            """\
            Specify period for index (overrides index default period).
            Must be one of annual,seasonal,monthly. For annual period,
            an optional argument can be given to specify a range of months,
            e.g. annual[jja] for multiple months, or annual[feb] for a
            single month.
            """
        ),
    )
    parser.add_argument(
        "-s",
        "--sliced-mode",
        action="store_true",
        help="activate calculation per period to avoid memory problems",
    )
    parser.add_argument(
        "--split-output",
        help=textwrap.dedent(
            """\
               Split the output into multiple files.

               Example of usage:
               --split-output month
               --split-output year
               --split-output season
               --split-output year[nr]

               Here 'nr' can be any number of year:
               --split-output year[5]
               --split-output year[10]
            """
        ),
    )
    parser.add_argument(
        "-o", "--output", dest="output_template", help="output filename"
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help=textwrap.dedent(
            """\
            By default, Climix will not overwrite an already existing output file.
            Use this switch to allow the overwriting of a file with the same
            output filename.
            """
        ),
    )
    parser.add_argument(
        "-f",
        "--metadata-file",
        metavar="METADATA_FILE",
        dest="metadata_files",
        action="append",
        help=textwrap.dedent(
            """\
            add an external metadata file (overrides any default definitions),
            can be used multiple times to add several files (the last specified
            file will override any earlier definitions)
            """
        ),
    )
    parser.add_argument(
        "--mask-start",
        metavar="MASK_START",
        dest="mask_start",
        action="append",
        help=textwrap.dedent(
            """\
            add a parameter file to mask the start of the data. Use to mask the start of
            the input data given a parameter-file. The grid of the input data and the
            parameter-file must be the same. The parameter-file should contain yearly
            data, where the data value should represent the day of the year. For each
            matching year all days before this day will be masked. If a year exists in
            the input data but not in the parameter-file this year will not be masked.
            This flag can be used for most standard indicies and works with any annual
            period, i.e, annual, annual[mjja] etc. If not a full year period is
            specified, a offset will be computed from the first day of the year and the
            first day of the data.
            """
        ),
    )
    parser.add_argument(
        "--mask-end",
        metavar="MASK_END",
        dest="mask_end",
        action="append",
        help=textwrap.dedent(
            """\
            add a parameter file to mask the end of the data. Use to mask the end of
            the input data given a parameter-file. The grid of the input data and the
            parameter-file must be the same. The parameter-file should contain yearly
            data, where the data value should represent the day of the year. For each
            matching year all days from this day until the end will be masked. If a year
            exists in the input data but not in the parameter-file this year will not
            be masked. This flag can be used for most standard indicies and works with
            any annual period, i.e, annual, annual[mjja] etc. If not a full year period
            is specified, a offset will be computed from the first day of the year and
            the first day of the data.
            """
        ),
    )
    parser.add_argument(
        "--activate-config",
        action="store_true",
        dest="activate_config",
        help=textwrap.dedent(
            """\
            specify if the climix configuration should be activated. This will apply
            the default configuration for the output global attributes metadata.
            The default configuration file can be overridden by adding an external
            climix config metadata file.
            """
        ),
    )
    parser.add_argument(
        "-r",
        "--reference-period",
        help=textwrap.dedent(
            """\
                specify reference period for an index (overrides the default
                reference period in the index definition), accepted formats are
                ISO 8601.
                Examples of usage:
                -r 1971/2000
                -r 1971-06-01/2000
                -r P48Y/2000
                -r P47Y7M/2000
                --reference-period 1971/2000
            """
        ),
    )
    parser.add_argument(
        "-cp",
        "--computational-period",
        help=textwrap.dedent(
            """\
                Specify computational period for the index, i.e. limit the time
                interval to a subset of what is available in the input data.
                The start and end time strings must follow `yyyy-mm-dd/yyyy-mm-dd`
                format, e.g. 1961-01-01/1990-12-31.
            """
        ),
    )
    parser.add_argument(
        "-x",
        "--index",
        action="append",
        required=True,
        metavar="INDEX",
        dest="indices",
        help=textwrap.dedent(
            """\
            the index to calculate
            (use "-x list" to get a list of all available indices)
            """
        ),
    )
    parser.add_argument(
        "datafiles", nargs="*", metavar="DATAFILE", help="the input data files"
    )
    return parser.parse_args()


def setup_logging(log_level, verbose=False):
    if verbose:
        format = (
            "%(relativeCreated)8dms:%(filename)s:%(funcName)s() "
            "%(levelname)s:%(name)s:%(message)s"
        )
    else:
        format = "%(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=LOG_LEVELS[log_level], format=format)


def setup_configuration(index, configuration):
    if configuration:
        index_distribution = index.metadata.distribution
        extra_attributes = {
            "CLIMIX_VERSION": f"climix-{__version__}",
            "INDEX_DISTRIBUTION": (
                f"{index_distribution.name}-{index_distribution.version}"
                if index_distribution is not None
                and index_distribution.name is not None
                else ""
            ),
        }
        configuration["global_attributes"].extra_attributes = extra_attributes


def build_output_filename(
    index, datafiles, output_template=None, computational_period=None
):
    """Construct output filename from template and basic information.

    Uses `output_template` to form the output filename by replacing the
    components `frequency` and `var_name`. If no `output_template` is given, it
    will try to guess a good template along CMIP/CORDEX naming standards.

    Parameters
    ----------
    index : climix.index.Index
        Index class.
    datafiles : list[str]
        List of files used to form output filename. Wildcards are expanded.
    output_template : str, optional
        A template to use as basis for output filename. If `None`, a template
        will be guessed based on the datafiles list.
    computational_period : str, optional
        A string with format `yyyy-mm-dd/yyyy-mm-dd` defining a time period.
        Default is None.

    Returns
    -------
    str
        Suggested output file name.
    """
    if output_template is None:
        path_list = []
        for file in datafiles:
            expand = [f for f in glob.glob(file)]
            path_list.extend(expand)

        if computational_period is not None:
            output_template = gordias.util.cmip_path.build_cmip_like_filename_template(
                path_list, time_range_placeholder=True
            )
            time_range = gordias.util.time_string.parse_isodate_time_range(
                computational_period
            )

            # Start and end time string should be in CORDEX/CMIP form, daily resolution.
            # Time range end is be inclusive, which means we have to reduce by a delta
            # (day resolution format = 8 characters) to get correct end CORDEX/CMIP time
            # string.
            #
            start_str = time_range.start.strftime("%Y%m%d")
            end_str = (
                time_range.end - gordias.util.time_string.TIME_RELATIVE_DELTAS[8]
            ).strftime("%Y%m%d")

            output_template = output_template.format(
                output_template.format,
                start=start_str,
                end=end_str,
                var_name="{var_name}",
                frequency="{frequency}",
            )
        else:
            output_template = gordias.util.cmip_path.build_cmip_like_filename_template(
                path_list
            )

    period_specialization = index.period.specialization
    if period_specialization is not None:
        var_name = f"{{var_name}}-{period_specialization}"
    else:
        var_name = "{var_name}"
    drs_copy = index.metadata.output.drs.copy()
    drs_copy["var_name"] = var_name.format(**index.metadata.output.drs)
    drs_copy["frequency"] = index.period.label
    return output_template.format(**drs_copy)


def do_main(
    index_catalog,
    configuration,
    requested_indices,
    period,
    reference_period,
    computational_period,
    datafiles,
    parameter_files,
    output_template,
    overwrite,
    sliced_mode,
    split_output,
    scheduler,
):
    logging.debug("Preparing indices")
    indices = index_catalog.prepare_indices(
        requested_indices, period, reference_period, computational_period
    )
    for index_no, index in enumerate(indices):
        logging.info(
            "Starting calculations for index <"
            f"{requested_indices[index_no]}> in {index}"
        )
        logging.debug("Building output filename")
        output_filename = build_output_filename(
            index, datafiles, output_template, computational_period
        )
        setup_configuration(index, configuration)
        logging.debug("Preparing input data")
        input_data = gordias.datahandling.prepare_input_data(datafiles, configuration)
        parameters = prepare_parameter_files(parameter_files, index)
        logging.debug("Calculating index")
        result = index(
            input_data,
            client=scheduler.client,
            sliced_mode=sliced_mode,
            parameters=parameters,
        )
        logging.info(f"Saving result in {os.path.abspath(output_filename)}")
        gordias.datahandling.save(
            result=result,
            output_filename=output_filename,
            split_output=split_output,
            client=scheduler.client,
            overwrite=overwrite,
            conventions_override=(configuration is not None),
            configuration=configuration,
        )


def prepare_parameter_files(parameter_files, index):
    """Prepare the parameter files used for masking the data.

    Parameters
    ----------
        parameter_files : list[str]
            Input parameter files to load.
        index :  climix.index.Index
            Index Class.

    Returns
    -------
        dict[str, iris.cube.Cube] | None
            A dict with the parameter cubes or `None` if no parameter files were given.

    """
    if parameter_files:
        assert isinstance(
            index.period, Annual
        ), "Parameter files is only supported for annual period."
        parameters = {}
        for key, value in parameter_files.items():
            params = gordias.datahandling.prepare_input_data(value)
            assert len(params) == 1, "Failed to concatenate parameter files."
            parameters[key] = next(iter(params))
    else:
        parameters = None
    return parameters


def parse_parameter_files(args):
    if args.mask_start is None and args.mask_end is None:
        return None
    parameter_files = {}
    if args.mask_start is not None:
        parameter_files["start"] = args.mask_start
    if args.mask_end is not None:
        parameter_files["end"] = args.mask_end
    return parameter_files


def main():
    args = parse_args()
    setup_logging(args.log_level, args.verbose)
    parameter_files = parse_parameter_files(args)
    logging.debug("Loading metadata")
    index_catalog, configuration = load_metadata(
        args.metadata_files, args.activate_config
    )
    if "list" in args.indices:
        print("Available indices are:")
        print(list(index_catalog.get_list()))
        return

    with gordias.dask_setup.setup_scheduler(args) as scheduler:
        logging.debug("Scheduler ready; starting main program.")
        start = time.time()
        try:
            do_main(
                index_catalog=index_catalog,
                configuration=configuration,
                requested_indices=args.indices,
                period=args.period,
                reference_period=args.reference_period,
                computational_period=args.computational_period,
                datafiles=args.datafiles,
                parameter_files=parameter_files,
                output_template=args.output_template,
                overwrite=args.overwrite,
                sliced_mode=args.sliced_mode,
                split_output=args.split_output,
                scheduler=scheduler,
            )
        finally:
            end = time.time()
            logging.info(f"Calculation took {end - start:.4f} seconds.")  # noqa: E231
        if args.keep_open:
            input("Press enter to close the cluster ")


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from importlib.metadata import version as get_version
from pathlib import Path

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
root = Path(__file__).absolute().parent.parent.parent
sys.path.insert(0, str(root))

# -- RTD configuration ------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from
# docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# This is used for linking and such so we link to the thing we're building
rtd_version = os.environ.get("READTHEDOCS_VERSION", "latest")
if rtd_version not in ["latest", "stable", "doc"]:
    rtd_version = "latest"


# -- Project information -----------------------------------------------------

project = "Climix"
copyright = "2019-2025, Lars Bärring, Klaus Zimmermann, Joakim Löw, Carolina Nilsson"
author = "Lars Bärring, Klaus Zimmermann, Joakim Löw, Carolina Nilsson"


release = get_version(project.lower())
version = release.split("+")[0]

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_copybutton",
    "sphinx_favicon",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "myst_parser",
    "sphinx_toolbox.collapse",
]

myst_enable_extensions = ["colon_fence", "substitution"]
myst_substitutions = {"version": version}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["climix.css"]
favicons = ["climixlogo-favicon.ico"]
html_theme_options = {
    "logo": {
        "image_light": "climixlogo.svg",
        "image_dark": "climixlogo_dark.svg",
    },
    "gitlab_url": "https://git.smhi.se/climix/climix",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

html_sidebars = {
    "**": [
        "version.html",
        "search-field.html",
        "sidebar-nav-bs.html",
        "sidebar-ethical-ads.html",
    ]
}


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
#
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
}



from cf_units import Unit
import dask.array as da
import numpy as np
import iris.coords

from .spell_kernels import make_first_spell_kernels, make_spell_length_kernels
from .support import (
    normalize_axis,
    IndexFunction,
    ThresholdMixin,
    ReducerMixin,
    DASK_OPERATORS,
)
from ..dask_take_along_axis import dask_take_along_axis

import gordias.dask_setup
import gordias.util.units

DASK_ARG_REDUCER = {
    "max": da.argmax,
    "min": da.argmin,
}


class FirstSpell(IndexFunction):
    def __init__(self, params):
        super().__init__(units=Unit("days"))
        self.params = params
        self.start_duration = self.params[0][2]
        if len(self.params) > 1:
            self.end_duration = self.params[1][2]
        else:
            self.end_duration = self.start_duration
        self.kernel = make_first_spell_kernels()

    def prepare(self, input_cubes, parameters=None):
        props = {
            (name, cube.dtype, cube.units, cube.standard_name)
            for name, cube in input_cubes.items()
        }
        for _, dtype, units, standard_name in props:
            for args in self.params:
                threshold = args[0]
                threshold.points = threshold.points.astype(dtype)
                if threshold.has_bounds():
                    threshold.bounds = threshold.bounds.astype(dtype)
                gordias.util.units.change_units(threshold, units, standard_name)
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        super().prepare(input_cubes)

    def pre_aggregate_shape(self, *args, **kwargs):
        return (4,)

    def call_func(self, data, axis, **kwargs):
        raise NotImplementedError

    def lazy_func(self, data, axis, cube, client, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        data = da.moveaxis(data, axis, -1)
        length = data.shape[-1]
        time = da.arange(length, dtype=np.float32)
        output_dim = data.ndim
        output_shape = 6  # length, head_chunk, res, tail_chunk, head_yr, tail_yr
        res_dim = output_dim + 1
        res_shape = 3  # start, end_independent, end_dependent
        data_idx = tuple(range(data.ndim))
        time_dim = data.ndim - 1
        time_idx = (time_dim,)
        output_idx = time_idx + data_idx[:-1] + (output_dim, res_dim)
        # divides data into chunks and looks for start and end spells for each chunk
        spell_res = da.blockwise(
            self.find_spell,
            output_idx,
            data,
            data_idx,
            time,
            time_idx,
            length=length,
            adjust_chunks={
                time_dim: lambda n: 1,
            },
            new_axes={
                output_dim: output_shape,
                res_dim: res_shape,
            },
            meta=np.array((), dtype=np.float32),
        )
        # reduces all chunks to one result by merging chunks in three steps
        res = client.persist(
            da.reduction(
                spell_res,
                self.chunk,
                self.aggregate,
                combine=self.combine,
                axis=0,
                keepdims=True,
                concatenate=True,
                split_every=4,
                dtype=np.float32,
                meta=np.array((), dtype=np.float32),
            )
        )
        gordias.dask_setup.progress(res)
        res = res[0]
        masked = da.ma.masked_array(
            da.ma.getdata(res),
            da.broadcast_to(mask[..., np.newaxis, np.newaxis], res.shape),
        )
        return masked[..., :2].astype(np.float32)

    def find_spell(self, chunk, time, length):
        """Calculate first spell information for a chunk of data.

        Parameters
        ----------
        chunk : da.array
            dask array containing one chunk of data
        time :  da.array
            dask array containing the time index for the chunked data
        length : int
            the length of the total data

        Returns
        -------
        da.array
            output dimensions : (1, n_long, n_lat, 6, 3)
            for each gridcell in the chunk returns 6x3 values:
            3 values:
            - start
            - end independent
            - end dependent
            6 values:
            - length of chunk
            - length of spell at beginning
            - position of first spell in the timeseries
            - length of spell at the end
            - length of spell at the beginning of the yr
            - length of the spell at the end of the yr
        """

        def _first_spell(chunk_data, chunk_time, args, leap_year=0):
            threshold, condition, duration, delay = args
            if delay > 59:
                delay += leap_year
            offset = np.where(delay <= chunk_time, True, False)
            thresholded_data = condition(chunk_data, threshold.points)
            thresholded_data = np.ma.filled(thresholded_data, fill_value=False)
            chunk_res = self.kernel.chunk(
                thresholded_data, offset, chunk_time, length, duration
            )
            return chunk_res

        stack = []
        leap_year = 0
        if length == 366:
            leap_year = 1
        # finds the independent start and end spells given conditions
        for args in self.params:
            chunk_res = _first_spell(chunk, time, args, leap_year)
            stack.append(chunk_res)
        # if there is no conditions given for the end
        if len(self.params) == 1:
            no_end = np.full(
                chunk.shape[:-1] + (6,), fill_value=np.nan, dtype=np.float32
            )
            stack.extend([no_end, no_end])
        else:
            end_dependent = stack[1].copy()
            start = stack[0][..., 2]
            cond = da.greater_equal(start, end_dependent[..., 2])
            for ind in np.ndindex(chunk.shape[:-1]):
                # if the start is before the end
                # compute a new dependent end with the start as offset
                if cond[ind]:
                    start_offset = int(time[0]) + start[ind] + 1
                    end_params = self.params[1][:-1] + (start_offset,)
                    chunk_res = _first_spell(chunk[ind], time, end_params)
                    end_dependent[ind] = chunk_res
            stack.append(end_dependent)
        res = da.stack(stack, axis=-1)
        return res.reshape((1,) + res.shape)

    def chunk(self, x_chunk, axis, keepdims, computing_meta=False):
        if computing_meta:
            return np.array((), dtype=np.float32)
        return x_chunk

    def combine(self, x_chunk, axis, keepdims):
        res = self.kernel.combine(
            np.array(x_chunk), self.start_duration, self.end_duration
        )
        return res.reshape((1,) + res.shape)

    def aggregate(self, x_chunk, axis, keepdims):
        res = self.kernel.combine(
            np.array(x_chunk), self.start_duration, self.end_duration
        )
        res = self.kernel.aggregate(res)
        return res.reshape((1,) + res.shape)

    def post_process(self, cube, chunk_data, coords, period, **kwargs):
        def _fuse(current_chunk, next_chunk):
            res = current_chunk.copy()
            for ind in np.ndindex(current_chunk.shape[:2]):
                ind_length = ind + (0, 0)
                ind_start_index = ind + (2, 0)
                ind_start_tail = ind + (5, 0)
                ind_end_index = ind + (2, 1)
                ind_end_tail = ind + (5, 1)
                ind_next_start_head = ind + (4, 0)
                ind_next_end_head = ind + (4, 1)
                # if no start have been found check if there is a start
                # at the end of the year
                if da.isnan(current_chunk[ind_start_index]):
                    start_spell = (
                        current_chunk[ind_start_tail] + next_chunk[ind_next_start_head]
                    )
                    # if there is a start
                    if start_spell >= self.start_duration:
                        res[ind_start_index] = (
                            current_chunk[ind_length]
                            - current_chunk[ind_start_tail]
                            + 1
                        )
                        end_tail = current_chunk[ind_end_tail]
                        # if the end tail is larger or equal to the start tail
                        # reduce the end tail
                        if current_chunk[ind_end_tail] >= current_chunk[ind_start_tail]:
                            end_tail = current_chunk[ind_start_tail] - 1
                        end_spell = end_tail + next_chunk[ind_next_end_head]
                        # if we have conditions for finding a end
                        if not da.isnan(end_spell):
                            # if there is a end after the start
                            if end_spell >= self.end_duration:
                                res[ind_end_index] = (
                                    current_chunk[ind_length] - end_tail + 1
                                )
                            # if the start is at the end of the year
                            # the end is not possible
                            elif res[ind_start_index] == current_chunk[ind_length]:
                                res[ind_end_index] = np.nan
                            # if there is no end found
                            # the end is set to be the end of the year
                            else:
                                res[ind_end_index] = current_chunk[ind_length]
                else:
                    # if there is a start but no end have been found
                    if da.isnan(current_chunk[ind_end_index]):
                        end_spell = (
                            current_chunk[ind_end_tail] + next_chunk[ind_next_end_head]
                        )
                        # if we have conditions for finding a end
                        if not da.isnan(end_spell):
                            # if there is a end
                            if end_spell >= self.end_duration:
                                res[ind_end_index] = (
                                    current_chunk[ind_length]
                                    - current_chunk[ind_end_tail]
                                    + 1
                                )
                            # if there is no end found
                            # the end is set to be the end of the year
                            else:
                                res[ind_end_index] = current_chunk[ind_length]
            return res

        mask = da.ma.getmaskarray(chunk_data)
        stack = []
        # start with the first year
        this = da.ma.getdata(chunk_data[(0,)]).rechunk(-1)
        tmp_res = da.empty(this.shape, dtype=np.float32)
        # add a year containing only zeros at the end for padding
        padding_chunk = da.zeros((1,) + this.shape, dtype=np.float32)
        padded_data = da.concatenate(
            [da.ma.getdata(chunk_data), padding_chunk], axis=0
        ).rechunk(-1)
        # for each year check if either the start or the end can be found in the overlap
        for next_chunk in padded_data[1:]:
            tmp_res = da.blockwise(
                _fuse,
                (0, 1, 2, 3),
                this,
                (0, 1, 2, 3),
                next_chunk,
                (0, 1, 2, 3),
                meta=np.array((), dtype=np.float32),
                align_arrays=False,
            )
            stack.append(tmp_res)
            this = next_chunk
        res_chunk = da.stack(stack, axis=0)
        masked_res = da.ma.masked_array(res_chunk, mask)
        return masked_res


class SeasonStart(FirstSpell):
    def __init__(self, **params):
        args = [
            (
                params["threshold"],
                DASK_OPERATORS[params["condition"]],
                params["duration"].points[0],
                params["delay"].points[0],
            )
        ]
        super().__init__(args)

    def lazy_func(self, data, axis, **kwargs):
        res = super().lazy_func(data, axis, **kwargs)
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        res = super().post_process(cube, data, coords, period, **kwargs)
        start = res[..., 0]
        return cube, start[..., 2].astype(np.float32)


class SeasonEnd(FirstSpell):
    def __init__(self, **params):
        args = [
            (
                params["start_threshold"],
                DASK_OPERATORS[params["start_condition"]],
                params["start_duration"].points[0],
                params["start_delay"].points[0],
            ),
            (
                params["end_threshold"],
                DASK_OPERATORS[params["end_condition"]],
                params["end_duration"].points[0],
                params["end_delay"].points[0],
            ),
        ]
        super().__init__(args)

    def lazy_func(self, data, axis, **kwargs):
        res = super().lazy_func(data, axis, **kwargs)
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        res = super().post_process(cube, data, coords, period, **kwargs)
        end = res[..., 1]
        return cube, end[..., 2].astype(np.float32)


class SeasonLength(FirstSpell):
    def __init__(self, **params):
        args = [
            (
                params["start_threshold"],
                DASK_OPERATORS[params["start_condition"]],
                params["start_duration"].points[0],
                params["start_delay"].points[0],
            ),
            (
                params["end_threshold"],
                DASK_OPERATORS[params["end_condition"]],
                params["end_duration"].points[0],
                params["end_delay"].points[0],
            ),
        ]
        super().__init__(args)

    def lazy_func(self, data, axis, **kwargs):
        res = super().lazy_func(data, axis, **kwargs)
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        res = super().post_process(cube, data, coords, period, **kwargs)
        start = res[..., 0]
        end = res[..., 1]
        length = end[..., 2] - start[..., 2] + 1
        length = da.ma.where(np.isnan(length), 0, length)
        return cube, length.astype(np.float32)


class StartOfSpring(SeasonStart):
    def __init__(self, **params):
        super().__init__(**params)
        self.leap_years = []

    def lazy_func(self, data, axis, **kwargs):
        res = super().lazy_func(data, axis, **kwargs)
        if data.shape[0] > 365:
            self.leap_years.append(1)
        else:
            self.leap_years.append(0)
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        _, res = super().post_process(cube, data, coords, period, **kwargs)
        start = res.copy()
        assert len(self.leap_years) > 0
        for year, leap in enumerate(self.leap_years):
            start[year] = da.where(res[year] > (212 + leap), np.nan, res[year])
        return cube, start.astype(np.float32)


class StartOfWinter(IndexFunction):
    def __init__(self, **params):
        super().__init__(units=Unit("days"))
        spring_args = {
            k.replace("spring_start_", ""): v
            for k, v in params.items()
            if "spring_start" in k
        }
        winter_args = {
            k.replace("winter_start_", ""): v
            for k, v in params.items()
            if "winter_start" in k
        }
        self.spring = StartOfSpring(**spring_args)
        self.winter = SeasonStart(**winter_args)
        self.leap_years = []

    def prepare(self, input_cubes, parameters=None):
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        self.spring.prepare(input_cubes, parameters)
        self.winter.prepare(input_cubes, parameters)
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name

    def pre_aggregate_shape(self, *args, **kwargs):
        return (4,)

    def call_func(self, data, axis, **kwargs):
        raise NotImplementedError

    def lazy_func(self, data, axis, **kwargs):
        spring_starts = self.spring.lazy_func(data, axis, **kwargs)
        winter_starts = self.winter.lazy_func(data, axis, **kwargs)
        res = da.stack([spring_starts, winter_starts], axis=-1)
        if data.shape[0] > 273:
            self.leap_years.append(1)
        else:
            self.leap_years.append(0)
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        _, spring_starts = self.spring.post_process(
            cube, data[..., 0], coords, period, **kwargs
        )
        _, winter_starts = self.winter.post_process(
            cube, data[..., 1], coords, period, **kwargs
        )
        start = winter_starts.copy()
        start = da.where(spring_starts > winter_starts, np.nan, start)
        assert len(self.leap_years) > 0
        for year, leap in enumerate(self.leap_years):
            start[year] = da.where(
                winter_starts[year] > (243 + leap), np.nan, winter_starts[year]
            )
        start -= 153
        return cube, start.astype(np.float32)


class StartOfSummer(IndexFunction):
    def __init__(self, **params):
        super().__init__(units=Unit("days"))
        spring_args = {
            k.replace("spring_start_", ""): v
            for k, v in params.items()
            if "spring_start" in k
        }
        summer_args = {
            k.replace("spring_", ""): v for k, v in params.items() if "spring" in k
        }
        autumn_args = {
            k.replace("autumn_start_", ""): v
            for k, v in params.items()
            if "autumn_start" in k
        }
        self.spring = StartOfSpring(**spring_args)
        self.summer = SeasonEnd(**summer_args)
        self.autumn = SeasonStart(**autumn_args)
        self.leap_years = []

    def prepare(self, input_cubes, parameters=None):
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        self.spring.prepare(input_cubes, parameters)
        self.summer.prepare(input_cubes, parameters)
        self.autumn.prepare(input_cubes, parameters)
        super().prepare(input_cubes)
        ref_cube = next(iter(input_cubes.values()))
        self.standard_name = ref_cube.standard_name

    def pre_aggregate_shape(self, *args, **kwargs):
        return (4,)

    def call_func(self, data, axis, **kwargs):
        raise NotImplementedError

    def lazy_func(self, data, axis, **kwargs):
        summer_starts = self.summer.lazy_func(data, axis, **kwargs)
        spring_starts = self.spring.lazy_func(data, axis, **kwargs)
        autumn_starts = self.autumn.lazy_func(data, axis, **kwargs)
        res = da.stack([summer_starts, spring_starts, autumn_starts], axis=-1)
        if data.shape[0] > 365:
            self.leap_years.append(1)
        else:
            self.leap_years.append(0)
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        _, summer_starts = self.summer.post_process(
            cube, data[..., 0], coords, period, **kwargs
        )
        _, spring_starts = self.spring.post_process(
            cube, data[..., 1], coords, period, **kwargs
        )
        _, autumn_starts = self.autumn.post_process(
            cube, data[..., 2], coords, period, **kwargs
        )
        start = summer_starts.copy()
        start = da.where(start >= autumn_starts, np.nan, summer_starts)
        start = da.where(np.isnan(spring_starts), np.nan, start)
        assert len(self.leap_years) > 0
        for year, leap in enumerate(self.leap_years):
            start[year] = da.where(start[year] > (304 + leap), np.nan, start[year])
        return cube, start.astype(np.float32)


class SpellLength(ThresholdMixin, ReducerMixin, IndexFunction):
    def __init__(self, threshold, condition, statistic, fuse_periods=False):
        super().__init__(threshold, condition, statistic, units=Unit("days"))
        self.spanning_spells = True
        self.kernels = make_spell_length_kernels(self.scalar_reducer)
        self.fuse_periods = fuse_periods
        self.lazy_arg_reducer = DASK_ARG_REDUCER[statistic]

    def prepare(self, input_cubes, parameters=None):
        if parameters is not None:
            raise NotImplementedError(
                "Parameter files cannot be used with this index function."
            )
        super().prepare(input_cubes)

    def pre_aggregate_shape(self, *args, **kwargs):
        return (4,)

    def call_func(self, data, axis, **kwargs):
        raise NotImplementedError

    def lazy_func(self, data, axis, **kwargs):
        axis = normalize_axis(axis, data.ndim)
        mask = da.ma.getmaskarray(data).any(axis=axis)
        data = da.moveaxis(data, axis, -1)
        res = da.reduction(
            data,
            self.chunk,
            self.aggregate,
            keepdims=True,
            output_size=7,
            axis=-1,
            dtype=np.float32,
            concatenate=False,
            meta=np.array((), dtype=np.float32),
        )
        res = da.ma.masked_array(
            da.ma.getdata(res), np.broadcast_to(mask[..., np.newaxis], res.shape)
        )
        return res.astype("float32")

    def chunk(self, raw_data, axis, keepdims, computing_meta=False):
        if computing_meta:
            return np.array((), dtype=int)

        data = self.condition(raw_data, self.threshold.points)
        data = np.ma.filled(data, fill_value=False)
        chunk_res = self.kernels.chunk(data)
        return chunk_res

    def aggregate(self, x_chunk, axis, keepdims):
        if not isinstance(x_chunk, list):
            return x_chunk
        res = self.kernels.aggregate(np.array(x_chunk))
        return res

    def post_process(self, cube, data, coords, period, **kwargs):
        def fuse(this, previous_tail):
            own_mask = da.ma.getmaskarray(this[..., 0])
            own_length = this[..., 0]
            own_head = this[..., 1]
            internal = this[..., 2]
            own_tail = this[..., 3]
            head = da.where(own_head, previous_tail + own_head, 0.0)
            tail = da.where(own_length == own_head, previous_tail + own_tail, own_tail)
            stack = da.stack([head, internal, tail], axis=-1)
            spell_length = da.ma.masked_array(
                self.lazy_reducer(stack, axis=-1), own_mask
            )
            return spell_length, tail

        if self.fuse_periods and len(data) > 1:
            stack = []
            this = data[0]
            slice_shape = this.shape[:-1]
            previous_tail = da.ma.masked_array(
                da.zeros(slice_shape, dtype=np.float32),
                da.ma.getmaskarray(data[0, ..., 3]),
            )

            for next_chunk in data[1:]:
                spell_length, previous_tail = fuse(this, previous_tail)
                stack.append(spell_length)
                this = next_chunk

            stack.append(fuse(next_chunk, previous_tail)[0])
            res_data = da.stack(stack, axis=0)
        else:
            res_data = self.lazy_reducer(data[..., 1:4], axis=-1)
            mask = da.ma.getmaskarray(res_data)
            res_ind = self.lazy_arg_reducer(data[..., 1:4], keepdims=True, axis=-1)
            spell_beginning = dask_take_along_axis(
                data[..., 4:], res_ind, axis=-1
            ).reshape(res_data.shape)
            masked_spell_beginning = da.ma.masked_array(spell_beginning, mask).squeeze()
            aux_coord = iris.coords.AuxCoord(
                masked_spell_beginning,
                var_name="spell_beginning",
                long_name="Day-of-year when longest spell begins",
                units=Unit("day"),
            )
            self.extra_coords.append(aux_coord)
        return cube, res_data
